(function (global) {
    "use strict";

    global.FRCC = global.FRCC || {};
    global.FRCC.Portal = global.FRCC.Portal || {};
    global.FRCC.Portal.Accessibility = (function () {
        /**
         * Global set of accessibility modifications to be made on the portal
         *
         * This will be executed from the EYGA Header Web Template so it is ran on every page
         */
        var loadGlobalAccessibilityModifications = function (configuration) {
            insertSpanElements(configuration.targetingObject);
            improveHeaderAccessibility();
            updateNativeTables();
            updateOfflineNotification();
            updateRestrictedPages(configuration.portalTitle);
        };

        /**
         * @param {object} targetingObject - key/value object containing selectors:insertMethod
         *
         * @example
         * insertSpanElements({
         *  'select': 'after',
         *  'input[type="radio"]': 'after',
         *  '.boolean-radio > label': 'prepend'
         * })
         *
         * This is an accessbility-specific function that was needed due to many fields generating without proper
         *      icon identification on entityforms within the portal.
         */
        var insertSpanElements = function (targetingObject) {
            // Cache default jQuery selector
            var $ = jQuery;

            // Native subgrids are generated within iframe elements on entity forms
            // Attaching a handler that will call the function to insert span elements when the iframe has completed its loading
            $("iframe[data-page]").on("load", function (e) {
                try {
                    _insertSpanElementsHelper(targetingObject, e.currentTarget.contentWindow.$);
                } catch (error) {
                    handleError(error);
                }
            });

            _insertSpanElementsHelper(targetingObject, $);
        };

        var _insertSpanElementsHelper = function (targetingObject, $) {
            if (typeof $ === "undefined" || $ === null) {
                return;
            }

            const spanElementHtml = '<span class="ui"></span>';

            for (const property in targetingObject) {
                var selector = property;
                var insertMethod = targetingObject[property];
                switch (insertMethod) {
                    case "after":
                        $(selector).after(spanElementHtml);
                        break;
                    case "prepend":
                        $(selector).prepend(spanElementHtml);
                        break;
                    default:
                        break;
                }
            }
        };

        /**
         * This function improves keyboard accessibility for the main navigation header.
         *
         * If the Escape key is pressed when the focus is within the header and a submenu is open, it removes the
         *      "in" class from the dropdown menu to close it.
         *
         * If the Up or Down keys are pressed and the focus is within the context of the navbar-links menu, it will
         *      mimic tab and shift-tab functionality
         */
        var improveHeaderAccessibility = function () {
            var $nav = $('nav[aria-label="main"]');
            var $navbarToggle = $(".navbar-toggle");
            var $webLinks = $(".weblinks");
            var $weblinkDropdown = $(".weblink.dropdown");

            document.addEventListener("keydown", function (e) {
                if (e.code === "Escape" && $nav.has(e.target).length && $nav.find('[aria-expanded="true"]').length) {
                    $("#navbar").removeClass("in");
                }

                if (
                    (e.code === "ArrowDown" || e.code === "ArrowUp") &&
                    ($navbarToggle[0] === e.target || $webLinks.has(e.target).length) &&
                    !$weblinkDropdown.has(e.target).length &&
                    $nav.find('[aria-expanded="true"]').length
                ) {
                    var currentElement = $(document.activeElement);
                    var nextElement = $(":tabbable").eq($(":tabbable").index(currentElement) + (e.code === "ArrowDown" ? 1 : -1));
                    if (!nextElement.hasClass("navbar-links") && e.code === "ArrowDown") {
                        e.preventDefault();
                        return;
                    }
                    nextElement.focus();
                    e.preventDefault();
                }
            });
        };

        /**
         * This function updates the classes used for native tables loaded onto the portal
         *
         * A MutationObserver is required because the native tables are loaded dynamically after
         *      the page/document is rendered
         *
         * table-fluid causes screen readers to not be able to find the table properly and needs to be replaced
         *
         * This function will replace the table-fluid class with table-mobile: a custom class that copies
         *      the styles used in table-fluid
         */
        var updateNativeTables = function () {
            // Native subgrids are generated within iframe elements on entity forms
            // Attaching a handler that will call the function to create a mutation observer when the iframe has finished loading
            $("iframe[data-page]").on("load", function (e) {
                try {
                    _registerMutationObserver(e.currentTarget.contentWindow.document, null, false, _updateNativeTablesHelper);
                } catch (error) {
                    handleError(error);
                }
            });

            _registerMutationObserver(document, "main", false, _updateNativeTablesHelper);

            _registerMutationObserver(document, ".view-pagination:not(.core-pagination-wrapper)", false, _updateNativePaginationHelper);
        };

        var _updateNativePaginationHelper = function (mutationsList) {
            mutationsList.forEach((mutation) => {
                if (mutation.target.classList.contains("view-pagination") && mutation.addedNodes.length > 0) {
                    var nativePaginationWrapper = $(".view-pagination:not(.core-pagination-wrapper)");

                    nativePaginationWrapper.find("ul.pagination").wrap("<nav aria-label='pagination'></nav>");

                    nativePaginationWrapper.find("a").each(function () {
                        $(this).removeAttr("role");

                        var ariaLabel = $(this).attr("aria-label");

                        if (ariaLabel == "Previous page" || ariaLabel == "Next page") {
                            var newTitle = $(this).attr("aria-label");
                            $(this).attr("title", newTitle);
                            $(this).removeAttr("data-toggle");
                            $(this).contents().wrap("<span aria-hidden='true'></span>");
                        } else if ($(this).attr("data-page") == "..") {
                            $(this).contents().wrap("<span aria-hidden='true'></span>");
                            $(this).attr("aria-label", "ellipsis indicating non-visible pages");
                        } else if (!$(this).parent().hasClass("active")) {
                            //if it's a number but not active
                            $(this).attr("aria-current", "false");
                        } else {
                            //number that is active
                            var pageNum = $(this).attr("data-page");
                            $(this).attr("aria-label", `page ${pageNum}`);
                        }

                        if ($(this).parent().hasClass("disabled")) {
                            $(this).removeAttr("href");
                            $(this).attr("aria-disabled", "true");
                        } else {
                            $(this).attr("href", "javascript:void(0)");
                        }
                    });
                }
            });
        };

        var _updateNativeTablesHelper = function (mutationsList) {
            mutationsList.forEach((mutation) => {
                mutation.addedNodes.forEach((node) => {
                    if (node.nodeName.toLowerCase() !== "table") {
                        return;
                    }

                    node.removeAttribute("aria-relevant"); // Remove aria-relevant="additions" - not supported by most browsers

                    const classNames = node.classList;
                    classNames.forEach((e) => {
                        if (!e.contains("table-fluid")) {
                            return;
                        }

                        var n = e.replace("fluid", "mobile");
                        classNames.replace(e, n);
                    });
                });
            });
        };

        /**
         * Set of accessibility modifications to be made on the portal against the native offlineNotificationBar
         */
        var updateOfflineNotification = function () {
            $("#offlineNotificationBar").attr("role", "alert");
            $("#web").attr("aria-hidden", "true");
            $("#close").attr({
                role: "button",
                tabindex: "0",
                style: "top:0.8rem",
            });
            $("#close").on("keypress", function (e) {
                if (e.key !== "Enter") {
                    return;
                }

                $(this).click();
            });
        };

        /**
         * Set of accessibility modifications to be made on the portal's "Restricted Pages"
         *
         * Restricted Pages are web pages where the HTML cannot be edited via a Web Template such as SignIn, Register, Redeem Invitation
         */
        var updateRestrictedPages = function (portalTitle) {
            const webpageTitle = window?.location?.pathname;
            if (!webpageTitle) {
                return;
            }

            const restrictedPages = ["/signin", "/register", "/account/login/register"];
            if (!restrictedPages.includes(webpageTitle.toLowerCase())) {
                return;
            }

            _updatePageTitles(portalTitle);
            _updatePageStructure(webpageTitle);
            _updateRestrictedPageAccessibility();
        };

        /**
         * Updates the restricted page's document.title attribute to be consistent with formatting across the Portal
         */
        var _updatePageTitles = function (portalTitle) {
            if (portalTitle == null) {
                return;
            }

            _registerMutationObserver(document, "title", false, (mutationsList) => {
                _updatePageTitle(mutationsList, portalTitle);
            });
        };

        var _updatePageTitle = function (mutationsList, portalTitle) {
            const documentTitle = document.title;
            const parts = documentTitle.split("-");
            const pageTitle = parts.slice(-1);
            if (pageTitle.length === 0) {
                return;
            }

            if (documentTitle === `${portalTitle} - ${pageTitle[0].trim()}`) {
                return;
            }

            document.title = `${portalTitle} - ${pageTitle[0]}`;
        };

        /**
         * Updates the restricted page's HTML structure to be consistent with structures used on other Web Pages throughout
         *      the portal
         */
        var _updatePageStructure = function (webpageTitle) {
            const pageTitle = webpageTitle.replace(/\//g, "").toLowerCase().trim();

            $("#content").addClass("body-content");
            $("#content").unwrap();
            $("#content").wrap(`<main id="content-container" class="page-content page-${pageTitle}" role="main"></main>`);
            $("#mainContent").removeClass("page-content");
            $("#mainContent").addClass("container");
            $(".nav.nav-tabs.nav-account").prependTo("#mainContent");
        };

        /**
         * Updates the restricted page's accessibility by removing redundant attributes and modifying validation alerts
         */
        var _updateRestrictedPageAccessibility = function () {
            _registerMutationObserver(document, ".form-horizontal", true, _modifyRestrictPageValidationAlerts);

            $(".nav.nav-tabs.nav-account > li > a").attr("tabindex", "0");
            $(".form-horizontal div.alert-danger").attr({
                role: "alert",
                "aria-live": "assertive",
            });
            $("*[title]").removeAttr("aria-label");
            $("*[title]").removeAttr("title");
        };

        var _modifyRestrictPageValidationAlerts = function (mutationsList) {
            $(".form-horizontal div.alert-danger li:visible a").each(function (index, ele) {
                var $ele = $(ele);
                var id = $ele.attr("href");
                $(id).closest(".form-group").addClass("has-error");
            });
        };

        var _registerMutationObserver = function (_document, querySelector, multiple, handler) {
            const observer = new MutationObserver(handler);
            const config = { childList: true, subtree: true };

            const targetNodes = multiple
                ? querySelector
                    ? _document.querySelectorAll(querySelector)
                    : [_document.body]
                : querySelector
                ? _document.querySelector(querySelector) !== null
                    ? [_document.querySelector(querySelector)]
                    : [_document.body]
                : [_document.body];

            targetNodes.forEach((node) => observer.observe(node, config));
        };

        /**
         * Application-specific set of accessibility modifications to be made on the portal
         *
         * This will be executed from the EYGA Templates: Application Create/Edit/View Web Templates
         */
        var loadApplicationAccessibilityModifications = function () {
            _removeRedundantAttributes();
            _modifyValidationAlerts();
            _modifyTableAlerts();
            _addPlaceholderText();
            _modifyProgressIndicatorModal();
            _removeRoleDescriptions();
            _modifyRequiredFieldLabels();
        };

        /**
         * Removes "aria-label" and "title" attributes from required fields on entityforms within the portal.
         *
         * This is an accessbility-specific function that was needed due to required fields generating
         *      with additional attributes that non-required fields were not generating with.
         * The aria-label and title attributes were causing screen readers to be read back the same information
         *      to the user, causing confusing and redundancy.
         */
        var _removeRedundantAttributes = function () {
            $('.required + .control > [aria-required="true"]').removeAttr("aria-label title");
            $(".required + .control > select").removeAttr("title");
            $('input[type="text"][required]').removeAttr("required").attr("aria-required", "true");
            $("input[readonly]").removeAttr("aria-label");
        };

        /**
         * Adds role="alert" to the validation summary elements on entityforms within the portal.
         *
         * This is an accessibility-specific function that was needed due to the validation areas on the form
         *      not having the proper role to inform screen reader users of the intended function of the element(s)
         *      on their screen.
         */
        var _modifyValidationAlerts = function () {
            var validationBlockSelector = ".validation-summary.alert.alert-error.alert-danger.alert-block";
            var validationBlock = $(validationBlockSelector);
            validationBlock.attr("role", "alert");

            _registerMutationObserver(document, validationBlockSelector, true, _addErrorClasses);

            $('input[value="x"]').val("×");
        };

        var _addErrorClasses = function (mutationsList) {
            mutationsList.forEach((mutation) => {
                $(mutation.target).closest(".entity-form").find(".has-error").removeClass("has-error");
                mutation.target.setAttribute("role", "alert");
                mutation.target.setAttribute("aria-live", "assertive");

                mutation.addedNodes.forEach((node) => {
                    if (node.nodeName.toLowerCase() !== "ul") {
                        return;
                    }

                    node.removeAttribute("role");
                    node.childNodes.forEach((childNode) => {
                        var href = childNode.childNodes[0].getAttribute("href").replace("#", "");
                        if (href.includes("_label")) {
                            $(`#${href}`).parent().siblings('.control').addClass('has-error');
                        } else if (href.includes("WebResource_FileUpload_")) {
                            document.getElementById(href).parentElement.previousElementSibling.classList.add("has-error");
                        } else {
                            $(`#${href}`).parent().addClass('has-error');
                        }
                    });
                });
            });
        };

        var _modifyTableAlerts = function () {
            $(".view-empty").removeAttr("role");
            $(".view-access-denied").attr("role", "alert");
            $(".view-error").attr("role", "alert");
        };

        /**
         * Adds placeholder text to select controls on entityforms within the portal
         *
         * This is an accessibility-specific function that was needed due to selects natively showing no placeholder
         *      text when no option was selected
         */
        var _addPlaceholderText = function () {
            $(".form-control.picklist option:first-child").text("Please select an option");
        };

        /**
         * Adds aria-atomic="true" and aria-live="assertive" to parent-level element of the progressIndicator element
         *
         * Since the Progress Indicator is loaded via a Content Snippet, it comes with extra baggage - 2 parent divs
         *      are loaded along with the HTML content of the Content Snippet. In order to correctly mark the element
         *      to screen readers as a "live" area, we have to do it at the parent-level
         */
        var _modifyProgressIndicatorModal = function () {
            var progressIndicator = $("#progressIndicator");
            if (progressIndicator.length === 0) {
                return;
            }

            progressIndicator.parent().parent().attr({ "aria-atomic": "true", "aria-live": "assertive" });
        };

        /**
         * Removes "aria-roledescription" from all inputs on entityforms within the portal.
         *
         * This is an accessbility-specific function that was needed due the aria-roledescription attribute
         *      overriding the selected attribute and reading the content of aria-roledescription to screen readers.
         */
        var _removeRoleDescriptions = function () {
            $("[aria-roledescription]").removeAttr("aria-roledescription");
        };

        /**
         * Adds <span class="sr-only">*</span> to all required input labels within the portal.
         *
         * This is an accessbility-specific function that was needed due to labels for required inputs
         *      not having proper acessibility for screen-readers.
         */
        var _modifyRequiredFieldLabels = function () {
            $(".info.required > label").each(function (e) {
                let labelHtml = $(this).html();
                $(this).html(`${labelHtml} <span class="sr-only">*</span>`);
            });
        };

        var handleError = function (errorMessage, displayError = false) {
            if (displayError) {
                alert(errorMessage);
            }

            console.error(errorMessage);
        };

        return {
            loadGlobalAccessibilityModifications: loadGlobalAccessibilityModifications,
            loadApplicationAccessibilityModifications: loadApplicationAccessibilityModifications,
        };
    })();
})(this);
